// $Id: WILL_Currency.js,v 1.16 2006/01/13 11:21:31 Will Exp $

//WILL: TODO: add detection for sepeator in arguements - warn
function WILL_Currency(country,description,code,displaySymbolIndex,symbols) {
  if (!arguments || !arguments.length || !(arguments.length==1 || arguments.length==5)) { throw "Require 1 string or 5 args to initalise Currency"; }
  if (arguments[0] && arguments[0]=='empty') { return this; } // for testbench

  var SEPERATOR='-';
  this.toString=toString;
  this.isValid=isValid;
  this.equals=equals;
  this._createFromString=_createFromString;
  this.getHtml=getHtml;
  this.symbols2str=symbols2str;
  this.str2symbols=str2symbols;
 
  this.displaySymbolIndex=displaySymbolIndex;

  if (arguments.length==1) {
    this._createFromString(arguments[0]); //the first arg is the complete string
  } else {
    this.country=country;
    this.description=description;
    this.code=code;
    symbols.unshift(code); //add code as the 1st element of the symbols array
    this.symbols=symbols;
  }
  this.error='';

  this.isValid();

  function getHtml() {
    if (!this.symbols || !this.symbols[this.displaySymbolIndex]) { throw 'Missing symbol'; } //missing symbol
    return this.symbols[this.displaySymbolIndex]; //First symbol is used for display, the rest for detecting in web pages
  }

  function _createFromString(s) {
    if (!s) { throw 'Tried to create a currency from null string'; }

    s=unescape(s);
    var bits=s.split(SEPERATOR);
    if (bits.length < 5) { throw 'Tried to create a currency from invalid string '+s; }
    this.country=bits.shift();
    this.description=bits.shift();
    this.code=bits.shift();
    this.displaySymbolIndex=bits.shift();
    this.symbols=bits; //all the rest
  }
      

  function toString(encoded) {
    var s;
    switch(encoded) {
    case true:
      // encode for storage
      s=this.country +SEPERATOR+ this.description +SEPERATOR+ this.code +SEPERATOR+ this.displaySymbolIndex +SEPERATOR+ this.symbols2str();
      s=escape(s);
      break;
    case false:
      // verbose mode for human
      s=this.code+' '+this.getHtml()+' - '+this.country + ' ' + this.description;
      break;
    default:
      //simple human readable form if toString() called
      s=this.code+' '+this.getHtml(); 
    }
    return s;
  }

  function symbols2str()    { return this.symbols.join(SEPERATOR); }
  function str2symbols(str) { this.symbols=str.split(SEPERATOR);   }

  function isValid() {
    //WILL: TODO: check for seperator
    var err='';
    if (!this.country || !this.description || !this.code || !this.symbols || !this.symbols.length) {
      WILL_log('ERR',this+' is missing entries');
      return false;
    }
    
    if (!this.code.length || this.code.length != 3) { 
      err += '\n Code "'+this.code+'" should be 3 letters (example: USD or EUR)';
    }
    //WILL: TODO: put validation back in
    //if (this.symbol.length < 1) err=err+'\n Symbol "'+this.symbol+'" should be at least 1 letter (example: $ or \\u02a3)';
    //if (this.html.length   < 1) err=err+'\n Html "'+this.html+'" should be at least 1 letter (example: $ or &euro;)';
    this.error=err;
    if (err) { return false; }
    else     { return true;  }
  }

  function equals(c) {
    if (this==c) { return true; }
    else if (!c) { return false; }
    else         { return (this.code == c.code); }
  }

  return this; // end WILL_Currency()
}

function WILL_isArray() {
  if (typeof arguments[0] == 'object') { 
    var criterion =  arguments[0].constructor.toString().match(/array/i); 
    return (criterion !== null); 
  }
  return false;
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
function WILL_CurrencyList(items) {
  var SEPERATOR=',';
  this.getNext        = getNext;
  //  this.findFromSymbol = findFromSymbol;
  this.findFromCode   = findFromCode;
  this.findFromCodes  = findFromCodes;
  this.findFromStr    = findFromStr;
  this.remove         = remove;
  this.add            = add;
  this.indexOf        = indexOf;
  this.existsInList   = existsInList;
  this.toString       = toString;
  this._createFromString = _createFromString;

  this._createRegExpIfMissing = _createRegExpIfMissing; //WILL: TODO: merge into _create
  this._createRegExps = _createRegExps;
  this._currencyRegExp   = null;
  this._currencyLeftRegExp   = null;
  this._currencyRightRegExp  = null;
  //  this._currencyEitherRegExp = null;

  this.getCurrencyRegExp  = getCurrencyRegExp;

  this.getCurrencyLeftRegExp  = getCurrencyLeftRegExp;
  this.getCurrencyRightRegExp = getCurrencyRightRegExp;

  this.getCurrencyLeftRegExp2  = getCurrencyLeftRegExp2;
  this.getCurrencyRightRegExp2 = getCurrencyRightRegExp2;


  this.isUnique       = isUnique;
  this.completelyUnique = true;
  this._checkAllUnique = _checkAllUnique;
  this._symbol2currency = {};
  this._sortedSymbolList = [];

  this.items=[];
  if (items) {
    if (WILL_isArray(items))   { this.items=items; }
    else                       { this._createFromString(items); }
  }
  this._checkAllUnique();

  function isUnique(str) {
    if (!this._symbol2currency[str])           { return true; } //WILL: should this happen?
    if (this._symbol2currency[str].length==1)  { return true; } // only one entry so it is unique
    return false;
  }
  
  function _checkAllUnique() {
    this.completelyUnique = true;
    this._symbol2currency={};
    this._sortedSymbolList = [];
    var currency;
    var symbol;
    for (var i=0;i<this.items.length;i++) {
      currency=this.items[i];
      
      for (var j=0;j<currency.symbols.length;j++) {
	symbol=currency.symbols[j].toLowerCase(); // lookup is always done on lower case symbol
	if (!this._symbol2currency[symbol]) {
	  this._symbol2currency[symbol]=[currency];
	} else {
	  this._symbol2currency[symbol].push(currency);
	  this.completelyUnique=false;
	  //WILL_log('DBG','Found non-unique '+this._symbol2currency[symbol]);
	}
      }
    }
    // regexps will need to be regenerated
    this._currencyRegExp   = null;
    this._currencyLeftRegExp   = null;
    this._currencyRightRegExp  = null;
    this._currencyLeftRegExp2   = null;
    this._currencyRightRegExp2  = null;
    //    this._currencyEitherRegExp = null;
  }

  function _createRegExpIfMissing(regexpName) {
    if (!this[regexpName]) { this._createRegExps();  }
    if (!this[regexpName]) { throw('unknown regexp name '+regexpName); }
    return this[regexpName];
  }
  function getCurrencyRegExp()       { return this._createRegExpIfMissing( '_currencyRegExp'       );  }
  function getCurrencyLeftRegExp()   { return this._createRegExpIfMissing( '_currencyLeftRegExp'   );  }
  function getCurrencyRightRegExp()  { return this._createRegExpIfMissing( '_currencyRightRegExp'  );  }
  function getCurrencyLeftRegExp2()  { return this._createRegExpIfMissing( '_currencyLeftRegExp2'  );  }
  function getCurrencyRightRegExp2() { return this._createRegExpIfMissing( '_currencyRightRegExp2' );  }


  function _createRegExps() {
    // now that we've built the associateArray, remember the keys (symbols) sorted by longest to shortest length
    // needed for regexp matching priority
    var sortByKeyLength = function(associateArray) {
      var keys=[];
      for (var k in associateArray) { keys.push(k); }
      keys.sort( function(a,b) { return (b.length-a.length); });
      return keys;
    };
    //this._sortedSymbolList //WILL: replace this funcationailty
    var symbolList=sortByKeyLength(this._symbol2currency);
    
    var matchCurr=symbolList.join('@@@'); //WILL: TODO: is this a unqiue seperator //change to verical tab
    //TEST: str.replace(/([.?|[\]\/^$*+\\{}()])/g , '--$1--');
    // Testing showed that only ] and / need to be escaped inside character class
    // but for emacs I needed to escape [ as well for syntax colouring

    // IMPORTANT - since single quotes are also interpolated in Javacript \b must be written '\\b'
    matchCurr = matchCurr.replace(/([.?|\[\]\/^$*+\\{}()])/g , '\\$1');			
    matchCurr = matchCurr.replace(/@@@/g, '|'); //now that we have escaped | in symbols, we can use it in our regexp
    matchCurr = '(' + matchCurr + ')';
    var matchNumber='((?:\\d[\\d, .]*\\d)|\\d)'; // multiple digit/SPACE/comma/point starting & ending in digit OR single digit (with less priority)
    var matchSpace='(\\s*)';      //optional

    var newCompiledRegexp = function(regexpStr,flags) {
      var regexp = new RegExp().compile(regexpStr, flags); // Multi-line, case-Insensitive, Global
      WILL_log('DBG3','complied regexp='+regexp.toSource());
      return regexp;
    };
    // \b needs the multi-line modifier to recognise a currency on a line-boundary. for example:  \n$1.40

    // WILL: cannot use \b - it doesn't consider the dollar $16 to be a word boundary
    // WILL TODO: should anything else be considered an boundary. For example DOT or COMMA?

    var matchBoundary = WILL_CurrencyList.matchBoundary;
    var matchStartOrBoundary = '('+matchBoundary+'|^)'; 
    var matchEndOrBoundary   = '('+matchBoundary+'|$)';
    this._currencyLeftRegExp   = newCompiledRegexp( matchStartOrBoundary + matchCurr   + matchSpace + matchNumber , 'mig');
    this._currencyRightRegExp  = newCompiledRegexp( matchNumber + matchSpace + matchCurr + matchEndOrBoundary     , 'mig');

    // WILL: DO we still need multi-line??
    this._currencyRegExp   = newCompiledRegexp( matchStartOrBoundary + matchCurr + matchEndOrBoundary , 'i'); //insensitive 

    // since the digits have already been found, they are anchored to that match using $ and ^
    //these forms don't include the number:
    this._currencyLeftRegExp2   = newCompiledRegexp( matchStartOrBoundary + matchCurr + matchSpace +'$' , 'i'); //not global
    this._currencyRightRegExp2  = newCompiledRegexp( '^' + matchSpace + matchCurr + matchEndOrBoundary   , 'i'); //only used for matching one currency

    //this._currencyEitherRegExp = newCompiledRegexp( matchCurr+'?' + matchSpace +  matchNumber + matchSpace + matchCurr+'?' );
				  
  }
    
  // no checking for duplicates
  function _createFromString(s) {
    this.items=[];
    var currlist_str=s.split(SEPERATOR);
    var curr;
    for (var i=0; i<currlist_str.length; ++i) {
      curr=new WILL_Currency(currlist_str[i]);
      this.items.push(curr);
    }
  }

  function toString(encode) {
    var s='';
    if (!this.items || !this.items.length) { return s; } //it's possible to have an empty list

    var sep_str=SEPERATOR;
    if (!encode) { sep_str +=' '; } //add a space if this is for humans

    for (var i=0; i<this.items.length-1;i++) {
      s=s+this.items[i].toString(encode)+sep_str;
    }
    s=s+this.items[i].toString(encode); //no comma at end
    return s;
  }

  function findFromCodes(codes) {
    var list=[];
    for (var i=0;i<codes.length;i++) {
      list.push(this.findFromCode(codes[i]));
    }
    return list;
  }
		
  function findFromCode(code,failSilent) {
    for (var i=0; i<this.items.length; ++i) {
      if (this.items[i].code==code) { return this.items[i]; }
    }
    if (!failSilent) { WILL_log('ERR','failed to find '+code+' in list'); }
    return null;
  }

  function findFromStr(str,silentFailure) {
    str=str.toLowerCase(); // symbols are stored in lowercase
    var currency=null;
    if (!this._symbol2currency) { WILL_alert('Currency list '+this+'\n has no _symbol2currency associateArray'); }

    var list=this._symbol2currency[str];
    if (!list || !list.length) {
      if (!silentFailure) { WILL_log('ERR','Failed to find currency ['+str+'] in list: '+this); }
      return null;
    }
    if (list.length>1) { WILL_alert('too many curriencies have '+str+' as a symbol. Just returning the first match'); }
    return list[0];
  }

  function indexOf(currency) {
    var array=this.items;
    if (!array) { return -1; }
    var index=-1;
    for (var i=0;i<array.length;i++) {
      if (array[i].equals(currency)) { index=i; }
    }
    return index;
  }
  
  function existsInList(currency) {
    if (this.indexOf(currency)==-1) { return false; }
    return true;
  }

  function add(currency,allowDups) {
    if (this.indexOf(currency) > -1 && !allowDups) { WILL_alert(currency+' is already in this list'); return false; }
    this.items.push(currency);
    this._checkAllUnique(); // could be optimised to just add only the new item
    return true;
  }

    
  function remove(currency) {
    var array=this.items;
    var delIndex=this.indexOf(currency);
      
    if (delIndex==-1) {WILL_alert('Failed to find '+currency+' in '+this); return false; }
    
    for (var i=delIndex;i<array.length-1;i++) {
      array[i]=array[i+1];
    }
    array.length--;

    this._symbol2currency={};
    this._checkAllUnique();
    return true;
  }

  // null is also returned if we get to the end. This should disable us
  function getNext(oldItem) {
    var index;

    if (!oldItem) {
      index=-1;
    } else {
      index=this.indexOf(oldItem);
      if (index==-1) {WILL_alert('Failed to find '+oldItem+' in '+this); return null; }
    }
    
    var nextItem=null;
    index++;
    if (index<this.items.length) { nextItem=this.items[index]; }
    
    return nextItem;
  }


}

WILL_CurrencyList.matchBoundary = '[\\s\\d\\(\\)<>=.,-]';

//WILL: TODO: see if I support http://www.alanwood.net/unicode/currency_symbols.html

function WILL_createCompleteCurrencyList() {
  var data=
    [
     ['America (USA)', 'Dollars', 'USD', '$','dollars','bucks','US $','US$'],
     ['Argentina', 'Pesos', 'ARS', '$'],
     ['Aruba', 'Guilders/Florins', 'AWG', '\u0192'],
     ['Australia', 'Dollars', 'AUD', 'A$', 'AU $'], // $ removed
     ['Bahamas', 'Dollars', 'BSD', '$'],
     ['Barbados', 'Dollars', 'BBD', '$'],
     ['Belize', 'Dollars', 'BZD', '$'],
     ['Bermuda', 'Dollars', 'BMD', '$'],
     ['Bolivia', 'Bolivianos', 'BOB', '$'],
     ['Brazil', 'Reais', 'BRL','\u0052$'],
     ['Britain (UK)', 'Pounds', 'GBP','\u00A3','pounds','pound','quid'], //WILL: TODO: check limits on number of symbols
     ['Brunei Darussalam', 'Dollars', 'BND','$'],
     ['Cambodia', 'Riels', 'KHR','\u17DB'],
     ['Canada', 'Dollars', 'CAD','C$','C $','CDN$'], //WILL: removed $ - clashes with USD
     ['Cayman Islands', 'Dollars', 'KYD','$'],
     ['Chile', 'Pesos', 'CLP','$'],
     ['China', 'Yuan Renminbi', 'CNY','\u5143','Yuan'],
     ['Colombia', 'Pesos', 'COP','\u20B1'],
     ['Costa Rica', 'Coln', 'CRC','\u20A1'],
     ['Cuba', 'Pesos', 'CUP','\u20B1'],
     ['Cyprus', 'Pounds', 'CYP','\u00A3'],
     ['Denmark', 'Kroner', 'DKK','\u0075\u0072'],
     ['Dominican Republic', 'Pesos', 'DOP','\u20B1'],
     ['East Caribbean', 'Dollars', 'XCD','$'],
     ['Egypt', 'Pounds', 'EGP','\u00A3'],
     ['El Salvador', 'Coln', 'SVC','\u20A1'],
     ['England (United Kingdom)', 'Pounds', 'GBP','\u00A3'],
     ['Europe', 'Euro','EUR', '\u20AC','euro','euros'],
     ['Falkland Islands', 'Pounds', 'FKP','\u00A3'],
     ['Fiji', 'Dollars', 'FJD','$'],
     ['Gibraltar', 'Pounds', 'GIP','\u00A3'],
     ['Guernsey', 'Pounds', 'GGP','\u00A3'],
     ['Guyana', 'Dollars', 'GYD','$'],
     ['Hong Kong', 'Dollars', 'HKD','$'],
     ['Hungary', 'Forint', 'HUF', 'Ft.','Ft'], 
     ['India', 'Rupees', 'INR','\u20A8','Rs.','Rs'], //extra entrys //first symbol is used for display, the rest for detecting in web pages
     ['Iran', 'Rials', 'IRR','\uFDFC'],
     ['Isle of Man', 'Pounds', 'IMP','\u00A3'],
     ['Israel', 'New Shekels', 'ILS','\u20AA'],
     ['Jamaica', 'Dollars', 'JMD','$'],
     ['Japan', 'Yen', 'JPY','\u00A5','yen','\u5186'],
     ['Jersey', 'Pounds', 'JEP','\u00A3'],
     ['Korea (North)', 'Won', 'KPW','\u20A9'],
     ['Korea (South)', 'Won', 'KRW','\u20A9'],
     ['Laos', 'Kips', 'LAK','\u20AD'],
     ['Lebanon', 'Pounds', 'LBP','\u00A3'],
     ['Liberia', 'Dollars', 'LRD','$'],
     ['Malta', 'Liri', 'MTL','\u20A4'],
     ['Mauritius', 'Rupees', 'MUR','\u20A8'],
     ['Mexico', 'Pesos', 'MXN','$'],
     ['Mongolia', 'Tugriks', 'MNT','\u20AE'],
     ['Namibia', 'Dollars', 'NAD','$'],
     ['Nepal', 'Rupees', 'NPR','\u20A8'],
     ['Netherlands Antilles', 'Guilders/Florins', 'ANG','\u0192'],
     ['New Zealand', 'Dollars', 'NZD','$'],
     ['Nigeria', 'Nairas', 'NGN','\u20A6'],
     ['North Korea', 'Won', 'KPW','\u20A9'],
     ['Oman', 'Rials', 'OMR','\ufDFC'],
     ['Pakistan', 'Rupees', 'PKR','\u20A8'],
     ['Peru', 'Nuevos Soles', 'PEN','\u0053\u002f\u002e'],
     ['Philippines', 'Pesos', 'PHP','\u20B1'],
     ['Qatar', 'Rials', 'QAR','\ufDFC'],
     ['Russia', 'Rubles', 'RUB','$'],
     ['Saint Helena', 'Pounds', 'SHP','\u00A3'],
     ['Saudi Arabia', 'Riyals', 'SAR','\ufDFC'],
     ['Seychelles', 'Rupees', 'SCR','\u20A8'],
     ['Singapore', 'Dollars', 'SGD','$'],
     ['Solomon Islands', 'Dollars', 'SBD','$'],
     ['South Africa', 'Rand', 'ZAR','R'],
     ['South Korea', 'Won', 'KRW','\u20A9'],
     ['Sri Lanka', 'Rupees', 'LKR','\u0BF9'],
     ['Sweden', 'Kronor', 'SEK','\u0075\u0072'],
     ['Switzerland','Franc','CHF','Fr.'],
     ['Suriname', 'Dollars', 'SRD','$'],
     ['Syria', 'Pounds', 'SYP','\u00A3'],
     ['Taiwan New', 'Dollars', 'TWD','$'],
     ['Thailand', 'Baht', 'THB','\u0E3F'],
     ['Trinidad and Tobago', 'Dollars', 'TTD','$'],
     ['Turkey', 'New Lira', 'TRY','\u20A4'],
     ['Turkey', 'Liras', 'TRL','\u20A4'],
     ['Tuvalu', 'Dollars', 'TVD','$'],
     ['United Kingdom', 'Pounds', 'GBP','\u00A3','pounds'],
     ['United States of America', 'Dollars', 'USD','$'],
     ['Uruguay', 'Pesos', 'UYU','\u20B1'],
     ['Vietnam', 'Dong', 'VND','\u20AB'],
     ['Yemen', 'Rials', 'YER','\uFDFC'],
     ['Zimbabwe', 'Dollars', 'ZWD','$']
     ];

  var bits;
  var curr;
  var items=[];
  
  for (var i=0; i<data.length; ++i) {
    bits=data[i];
    var country=bits.shift();
    var description=bits.shift();
    var code=bits.shift();
    var symbols=bits; //all the rest
    var dspSymbolIndex=1;
    curr = new WILL_Currency(country,description,code,dspSymbolIndex,symbols);
    items.push(curr);
  }
  var currList = new WILL_CurrencyList(items);
  return currList;
}



// Ununsed Unicode conversion

function WILL_escapeUnicode(str) {
  var newstr='';
  var code,hex,newchr;
  for (var i=0;i<str.length;i++) {
    code=str.charCodeAt(i);
    if (code<128) {
      newchr=str[i];
    } else {
      hex=code.toString(16);
      if      (code<0x100) { newchr='\\u00'+hex; }
      else if (code<0x1000) { newchr='\\u0'+hex; }
      else if (code<0x10000) { newchr='\\u'+hex; }
      else                   { WILL_alert('Internal error - invalid unicode '+code); }
    }
    newstr=newstr+newchr;
  }
  return newstr;
}
