// $Id: WILL_ExchangeRateCache.js,v 1.12 2006/01/13 23:08:33 Will Exp $

//WILL: should the cache just watch the relevant prefs?

// An exchange rate can have:
// SOURCE:
// * web(1)     - from 1st web service
// * fixed      - fixed value e.g. 1.51 //WILL: TODO:
// * calculated - calculated from another exchange rate, e.g. 40*EUR //WILL: TODO

function WILL_ExchangeRate(fromCurrency,toCurrency,prefGetWebServiceUrlFunc) {
  var STATUS_OK      = "OK";      // present, well-formed and not old
  var STATUS_OLD     = "OLD";     // OK but needs web-update
  var STATUS_INVALID = "INVALID"; // missing, or not well-formed (0)

  var DEFAULT_SOURCE = "WEB(1)";
  var SEPERATOR      = '-';

  this.fromCurrency  = fromCurrency;
  this.toCurrency    = toCurrency;
  this.rate          = NaN;
  this.status        = STATUS_INVALID; //until my rate is set
  this.source        = DEFAULT_SOURCE;
  this.lastGoodUpdateTime = null;
  this.lastUpdateErr  = null;
  this._prefGetWebServiceUrlFunc = prefGetWebServiceUrlFunc; // used to get the base URL for the web service

//this.toString = function()     { return toCurrency.code+'/'+fromCurrency.code+' ('+this.status+')' };

  this.toString = function(encoded) {
    if (!encoded) {
      return this.fromCurrency.code+SEPERATOR+this.toCurrency.code; //human readable
    } else {
      return [this.toCurrency.code, 
	      this.fromCurrency.code,
	      this.rate,
	      this.status,
	      this.source,
	      this.lastGoodUpdateTime.valueOf() ].join(SEPERATOR); //datetime in ms
   }
  };
  this.updateFromStr = function(str) {
    var bits=str.split(SEPERATOR);
    if (bits.length != 6)                       { WILL_log('ERR','malformed rate: not 6 bits in str ['+str+']'); return; }
    if (this.toCurrency.code != bits.shift())   { WILL_log('ERR','malformed rate: missing To Code in str ['+str+']'); return; } //WILL: does it really make sense to store these twice?
    if (this.fromCurrency.code != bits.shift()) { WILL_log('ERR','malformed rate: missing From code in str ['+str+']'); return; }
    this.rate = bits.shift();
    this.status = bits.shift();
    this.source = bits.shift();
    this.lastGoodUpdateTime = new Date(Number(bits.shift()));
  };

  this.isStatusInvalid  = function()    { return this.status==STATUS_INVALID; };
  this.isStatusOld      = function()    { return this.status==STATUS_OLD; };
  this.isStatusOK       = function()    { return this.status==STATUS_OK; };
  this.setStatusOK      = function()    { this.status=STATUS_OK; this.lastUpdateErr  = null; };
  this.setStatusInvalid = function()    { this.status=STATUS_INVALID; };
  this.setStatusOld     = function()    { this.status=STATUS_OLD; };

  this.setRate  = function(rate) {
    this.rate=rate;
    if (rate) { this.status=STATUS_OK;      }
    else      { this.status=STATUS_INVALID; }
  };

  this.removeUpdateError = function() { this.lastUpdateErr=null; };

  this.setWebUpdate = function(success,err) {
    // this just records if the upate was successful or not
    // the status will be changed by setRate or by WILL: TODO: checkNotOld
    if (success) { this.lastGoodUpdateTime = new Date(); }
    else         { this.lastUpdateErr=err;               }
  };
  this.getRate = function() {
    if (this.status == STATUS_INVALID) { WILL_log('INFO','tried to use invalid ex rate '+this); }
    return this.rate;
  }; 
  this.getLastGoodUpdateTime = function() { 
    if (!this.lastGoodUpdateTime) { return 'Never successfully updated'; }
    else                          { return this.lastGoodUpdateTime.toLocaleString(); }
  }; 
  
  // sets STATUS_OLD and returns true is we haven't been updated in X hours
  this.setOldIfLastUpdateInHours = function(ageLimit) {
    if (!this.lastGoodUpdateTime) {this.setStatusOld(); return true;}

    var WILL_ageInHours = function(pastTime) { //WILL: Do we actually need this anywhere else?
      var oldTime = pastTime.getTime();
      var nowTime = new Date().getTime();
      var ageInHours= (( nowTime - oldTime ) / (1000*60*60)) ;
      return ageInHours;
    };    

    if (WILL_ageInHours(this.lastGoodUpdateTime) > ageLimit ) {
      this.setStatusOld();
      return true;
    }
  };

  this.getUpdateUrl = function() {
    var origUrl;
    var url=this._prefGetWebServiceUrlFunc(1); //get the first (default) webservice# //WILL: TODO: remove hard coded 1

    origUrl=url;
    url=url.replace(/FROM_CODE/, fromCurrency.code);
    if (origUrl==url) { WILL_alert('could not set FROM_CODE in '+url); }

    origUrl=url;
    url=url.replace(/TO_CODE/,   toCurrency.code);
    if (origUrl==url) { WILL_alert('could not set TO_CODE in '+url); }

    return url;

  };

}

// WILL: cache is currenly only used for getUrl and setRate
// WILL: items are rateObj's
// Needs: 
// * setFromXmlCallback(item,xml) which returns success/failure
// * setFailureCallback(item,errorStr);
// * list of items
//   - item.getUpdateUrl()

function WILL_WebUpdater() {
  var self=this;
  //public funcs
  this.start = start;
  this.numItemsToDo = function() { return this._itemsToDo.length; };
  this.fractionLeft = function() { return this._itemsToDo.length / this._origItemCount; };
  //private funcs
  this._startNextUpdate = _startNextUpdate;
  this._onLoadXML      = _onLoadXML;

  //private vars
  this._setFromXmlCallback = null;
  this._setFailureCallback = null;

  this.isBusy       = false; // WILL: TODO: make this gettable only from outside
  this._itemsToDo   = [];    // list of rates to be downloaded from web
  this._origItemCount = 0;   // how many items were we asked to do
  this._hadErrors   = false; // was there an error?
  this._currentItem = null;  // the currently downloading rate
  
  function start(itemsToDo,setFromXmlCallback,setFailureCallback) {
    WILL_log('INFO','  ------------ Web update starting ------------ ');
    if (this.isBusy) { WILL_alert('WebUpdater is already busy'); }
    this._setFromXmlCallback = setFromXmlCallback;
    this._setFailureCallback = setFailureCallback;

    this._itemsToDo=[];
    for (var i=0; i<itemsToDo.length; i++) { this._itemsToDo[i] = itemsToDo[i]; } //duplicate the array
    this._origItemCount = itemsToDo.length;
    //WILL: is this logic needed? for (var i=0; i<this._itemsToDo.length; i++) { this._itemsToDo[i].removeUpdateError(); }
    this.isBusy=true;
    this._notify('START');
    this._hadErrors=false;
    //kick-off the cycle _startNextUpdate --> (request.send) --> _onLoadXML --> back to _startNextUpdate()
    setTimeout( function() { self._startNextUpdate.call(self); } ,  100 ); //wait 10th of a sec and then start in a new thread
  }

  function _startNextUpdate() {
    if (this._itemsToDo.length===0) {
      if (this._hadErrors) {
	this._notify('END_ERRORS');
	WILL_log('INFO','  -----Web update complete - HAD ERRORS ! ----- ');
      } else {
	WILL_log('INFO','  ---------- Web update complete OK ----------- ');
	this._notify('END_OK');
      }
      this.isBusy=false; //mark the updater as finished
      return; //don't start any more requests
    }

    var itemForUpdate = this._itemsToDo.shift(); //get the next item to process
    this._currentItem=itemForUpdate;

    var url=itemForUpdate.getUpdateUrl();
    if (!url) { WILL_alert(itemForUpdate+' missing update URL'); }

    WILL_log('INFO','Updating '+itemForUpdate+' from: '+url);
  
    var request = new XMLHttpRequest(); 
    // You can sometimes add  your own properties to an object with Javascript - great! But not here :-(
    //request.WILL_rateObj=rateObj;
    request.onload  =  function(evt) { self._onLoadXML.call(self,evt); };
    request.onerror =  function(evt) { self._onLoadXML.call(self,evt); }; //WILL: what kind of event is passed to onLoad?

    try {
      request.open("GET", url, true); // async
      //dump(url+'\n');  // to monitor web access
      request.send(null); //null body
    } catch (e) {
      this._onLoadXML(e); // use the same handler even if an exception is immediately raised
    }
  }

  //we assume that this is the request that came from the last xmlHttpRequest
  function _onLoadXML(event) {

    // helper func
    var summarizeError = function() {
      var errorStr;
      errorStr=request.statusText+' ('+request.status+') ';
      errorStr=errorStr + request.responseText.replace(/<[^>]*?>/g,'');  //remove all tags
      errorStr=errorStr.replace(/\n\n+/g,'\n'); //remove multiple newlines
      return errorStr;
    };

    try {
      if (!event)        { throw "Missing event!"; }
      if (event.name)    { throw "Got an immediate exception - Failure: "+event.name; } 
      if (!event.target) { throw "Missing target"; }
      if (event.type == 'error') { throw 'Page failed to load'; }

      var request=event.target;

      if (request.status === undefined) { throw "Missing status"; }
      if (request.status != 200 && request.status !== 0) { throw "not status 200 OK or 0 (file load)"+summarizeError(); }

      var errorMsg = this._setFromXmlCallback(this._currentItem,request.responseText);
      if (errorMsg) { throw errorMsg; }

      // else we successfully updated
      WILL_log('INFO','Updated OK from web: '+this._currentItem); //WILL: TODO: move log to callback
      WILL_log('DBG',request.responseText);
      //dump(request.responseText+'\n\n'); // to monitor web access
      this._notify('RATE_UPDATED');
      
    } catch(e) {
      WILL_log('WARN','WebUpdate caught an exception: '+e);
      this._setFailureCallback(this._currentItem,e);
      this._notify('RATE_ERROR'); //report each time
      this._hadErrors=true;
    }
    this._currentItem=null; 
    this._startNextUpdate() ;    //start next request
  }

  // WILL: TODO: make these inhertied  //////////////////////////////////////

  this._observers = []; //we call these when download starts / stops / has error

  // WILL: If I need to make this async then consider http://devedge-temp.mozilla.org/toolbox/examples/2003/CCallWrapper/index_en.html
  this._notify = function(arg)  {
    WILL_log('DBG','Notify '+arg);
    for (var i = 0; i < this._observers.length; ++i) {
      var observer=this._observers[i];
      //WILL: This doesn't work. observer and i take their last values
      //WILL: setTimeout( function() { observer.observe(arg) } , 10 ); // call each objects obvserve method after 10ms
      observer.observe(arg); // make blocking calls
    }
  };
  
  this.addObserver = function (aObserver) {
    if (!aObserver.observe) { WILL_alert('addObserver: not an observer'); }
    this._observers.push(aObserver);
  };

  this.removeObserver = function (aObserver) {
    for (var i = 0; i < this._observers.length; ++i) {
      if (this._observers[i]==aObserver) {
	this._observers.splice(i,1); //remove this one item
	return;
      }
    }
    WILL_log('WARN','WILL: could not find '+aObserver+' in observer list. This could be because window was closed before observer had a chance to register');
  };

  ///////// WILL: end of observable /////////////////////////////////////
}


 function WILL_ExchangeRateCache(prefReadRateFunc, prefWriteRateFunc, prefGetWebServiceUrlFunc, sourceCurrencyList, targetCurrencyList) {
   var self=this;

  //This is just a cached copy of values stored in GM preferences. Access using getRate
  this.exchangeRates = {};
  this._prefReadRateFunc         = prefReadRateFunc;
  this._prefWriteRateFunc        = prefWriteRateFunc;
  this._prefGetWebServiceUrlFunc = prefGetWebServiceUrlFunc;

  this.webUpdater = new WILL_WebUpdater(this);

  // functions //
  this.setSourceAndTargetLists = setSourceAndTargetLists;
  this.getRate              = getRate;
  this.getRateObj           = getRateObj;
  this.setAllRatesFromPrefs = setAllRatesFromPrefs;
  this.updateFromWeb        = updateFromWeb;
 
  // private functions //
  this._checkForOldRates=_checkForOldRates;

  this._getKey=_getKey;
  this._setRate=_setRate;
  this._getAllRateObjs=_getAllRateObjs;

  this._setFromXmlCallback = _setFromXmlCallback; // used by the WebUpdater
  this._setFailureCallback = _setFailureCallback;

  //private vars
  this.missingRates=[]; //key is provided by _getKey


  var UPDATE_INTERVAL=8; //in hours

  this.setSourceAndTargetLists(sourceCurrencyList,targetCurrencyList);

  
  function setSourceAndTargetLists(sourceCurrencyList,targetCurrencyList) {
    this.sourceCurrencyList=sourceCurrencyList; 
    this.targetCurrencyList=targetCurrencyList;
    this.setAllRatesFromPrefs();
    this._checkForOldRates();
  }
    

  // this won't work until the rates have been updated from prefs
  function getRate(fromCurrency,toCurrency) {
    if (fromCurrency.equals(toCurrency)) { return 1; } //we don't bother with Exchange Rate objects for the obvious!

    var ex=this.getRateObj(fromCurrency,toCurrency);
    return ex.getRate();
  }
  function getRateObj(fromCurrency,toCurrency, createMissing) {
    if (fromCurrency.equals(toCurrency)) { return 1; } //NB this is not a rateObj!
    var key = toCurrency.code+'/'+fromCurrency.code; //WILL: check - we aren't using _getKey because we don't have a rateObj yet. Is this sensible?
    var ex=this.exchangeRates[key];
    if (!ex) {
      if (createMissing) {
	ex = new WILL_ExchangeRate(fromCurrency,toCurrency,this._prefGetWebServiceUrlFunc);
	this.exchangeRates[key]=ex;
      } else {
	//we weren't told to create missing objects - so return null
	WILL_log('DBG','could not find rate '+key);
	return null;
      }
    }
    return ex;
  }

  //////////////////////////////////////////////////////////////////////////
  // All the rest of the code is for keeping the exchange rate up to date //
  //////////////////////////////////////////////////////////////////////////

  function _getAllRateObjs() {
    var target_list=this.targetCurrencyList.items;
    var source_list=this.sourceCurrencyList.items;
    var toCurrency, fromCurrency;
    var rates=[];
    var rateObj;

    for (var t=0; t<target_list.length; t++) {
      toCurrency=target_list[t];
      for (var v=0; v<source_list.length; v++) {
	fromCurrency=source_list[v];
	if (fromCurrency.equals(toCurrency)) { continue; }//we don't bother with Exchange Rate objects for the obvious!
	rateObj=this.getRateObj(fromCurrency,toCurrency,true); //create any missing objs
	rates.push(rateObj);
      }
    }
    return rates;
  }

  function _checkForOldRates() {
    var oldRates=[];
    var allRates=this._getAllRateObjs();
    for (var i=0; i<allRates.length; i++) { 
      if (allRates[i].setOldIfLastUpdateInHours(UPDATE_INTERVAL)) {
	oldRates.push(allRates[i]);
      }
    }
    return oldRates;
  }

  function setAllRatesFromPrefs() {
    this.exchangeRates=[]; //empty the hashtable
    var allRates=this._getAllRateObjs();
    var key;
    var rateObj;
    for (var i=0; i<allRates.length; i++) {
      rateObj = allRates[i];
      key = this._getKey(rateObj);
      this._prefReadRateFunc(rateObj); // update from prefs
      this.exchangeRates[key]=rateObj; // update the hash
    }
  }

  //////////////////////////
  //private from here on: //
  //////////////////////////


  function _getKey(rateObj) {
    if (!rateObj) { WILL_alert('Missing rate obj'); }
    return rateObj.toCurrency.code+'/'+rateObj.fromCurrency.code;
  }

  function _setRate(rateObj,rate) {
    var key = this._getKey(rateObj);
    var ex=this.exchangeRates[key];
    if (!ex) { WILL_alert('tried to get missing exchange rate '+key); }
    ex.setRate(rate); //update the cached copy
    this._prefWriteRateFunc(rateObj);  //update the pref copy (this is a write-through cache)
    WILL_log('DBG3','Set rate: '+key+' = '+rate);
  }

  
  // returns errorStr if there was a problem
  function _setFromXmlCallback(item,xml) {
    // alert('got '+item+' xml='+xml);

    try {
      var parser    = new DOMParser();
      var xmlDoc    = parser.parseFromString(xml,"application/xml");
      if (!xmlDoc) { throw 'Could not create DOM from '+xml; }

      var rate_text = xmlDoc.getElementsByTagName('double')[0];
      if (rate_text===undefined) { throw 'Could not get rate for '+item+' in '+xml; }

      // we found the XML double tag
      var rate=rate_text.textContent;
      if (!rate || rate=='0') { throw 'Received unlikely exchange rate = '+rate+'\n'+xml; }

      // SUCCESS!
      item.setWebUpdate(true);
      self._setRate(item,rate);

    } catch (e) {
      return e.toString();
    }
    return undefined; // no error
  }
  function _setFailureCallback(item,errorStr) {
    item.setWebUpdate(false,errorStr);
    // alert('got failure for '+item+' error='+errorStr);
  }

  function updateFromWeb(forceUpdate) {
    if (this.webUpdater.isBusy) {
      WILL_log('WARN','Webupdater is already busy - not starting');
      return;
    }
    var todo;
    if (forceUpdate) { todo = this._getAllRateObjs();   }
    else             { todo = this._checkForOldRates(); }
    //   foo();
    if (todo.length) { this.webUpdater.start(todo,this._setFromXmlCallback,this._setFailureCallback); }
    else             { WILL_log('INFO','Web update skipped - no rates require update'); }
  }



}

