// $Id: WILL_NumberConversion.js,v 1.18 2006/01/11 13:56:06 Will Exp $

function WILL_findLongestNumber(textStr) {
  // textStr should start with a number hence leading ^
  var thousandsSpace_OptionalDecimal_RE = /^\d{1,3}(?:\s\d\d\d)*(?:[,.]\d+)?/; // 1-3 digits (space 3 digits)* (comma|dot digits+)?
  var thousandsComma_OptionalDecimal_RE = /^\d{1,3}(?:\,\d\d\d)*(?:\.\d+)?/;   // 1-3 digits (comma 3 digits)* (dot       digits+)?
  var thousandsPoint_OptionalDecimal_RE = /^\d{1,3}(?:\.\d\d\d)*(?:\,\d+)?/;   // 1-3 digits (point 3 digits)* (comma     digits+)?
  var thousandsQuote_OptionalDecimal_RE = /^\d{1,3}(?:\'\d\d\d)*(?:\.\d+)?/;   // 1-3 digits (quote 3 digits)* (dot       digits+)?
  var thousandsPlain_OptionalDecimal_RE = /^\d+(?:[,.]\d+)?/;                  // 1 or more digits (comma|dot digits+)?

  var r;
  var regexps = [ thousandsSpace_OptionalDecimal_RE,
		  thousandsComma_OptionalDecimal_RE,
		  thousandsPoint_OptionalDecimal_RE,
		  thousandsQuote_OptionalDecimal_RE,
		  thousandsPlain_OptionalDecimal_RE ];
    
  var numberStr='';
  for (var i=0;i<regexps.length;++i) {
    r=regexps[i].exec(textStr);
    if (r && r[0].length > numberStr.length) {
      numberStr=r[0];
    }
  }
  
  if (!numberStr) { throw 'Cannot create number from '+textStr; }
  //WILL_log('DBG','Found number ['+numberStr+']');
  return numberStr;
}


function WILL_Number(text) {
  if (arguments[0] && arguments[0]=='empty') { return this; } // for testbench

  // private vars (actually public)
  this._origStr = null;     // string representation of this number (in original format)
  this._origLen = 0;        // and length of that string
  this._origNum = NaN;      // float  representation of this number
  this._origNumSigFigs=NaN; // Original number of significant figures
  this._origNumSigDecs=NaN; // Original number of significant decimals (-ve number) //WILL: is this used??

  this._newNum  = NaN;      // either the orignal number or the original number multiplied by a rate

  //this._decimalSeperator=null; WILL: not used: 
  this._thousandsSeperator=null;

  //private vars
  var self=this;
  
  var POINT_RE=/\./g;
  var COMMA_RE=/\,/g;
  var SPACE_RE=/\s/g;
  var QUOTE_RE=/\'/g;

  // class vars //WILL: TODO: can I make this class variables. Even with my testbench?
  this.POINT='.';
  this.COMMA=',';
  this.SPACE=' ';
  this.QUOTE="'";
  this.BLANK='';

    
  _init(text);

  // private funcs
  function _init(text) {

    var number = WILL_findLongestNumber(text);

    self._origStr = number;
    self._origLen = number.length;

    _parseString(number);
    // Sets the following:
    // self._origNum and self._newNum
    // self._origNumSigFigs
    // self._origNumSigDecs
  }

  // convert into a Javascript formatted number and store as float
  function _parseString(numStr) {

    detectSeperators(numStr);

    // remove thousands seperators and use POINT as decimal seperator
    switch (self._thousandsSeperator) {
    case self.BLANK:
      break;
    case self.COMMA:
      numStr = numStr.replace(COMMA_RE,'');
      break;
    case self.SPACE:
      numStr = numStr.replace(SPACE_RE,'');
      break;
    case self.QUOTE:
      numStr = numStr.replace(QUOTE_RE,'');
      break;
    case self.POINT:
      numStr = numStr.replace(POINT_RE,'');
      break;
    default:
      throw 'Unknown thousands seperator '+self._thousandsSeperator;
    }

    switch (self._decimalSeperator) {
    case self.COMMA:
      numStr = numStr.replace(COMMA_RE,self.POINT); // POINT as decimal
      break;
    case self.POINT:      
      break; //do nothing
    default:
      throw 'Unknown decimal seperator '+self._decimalSeperator;
    }

    self._origNum=parseFloat(numStr);
    self._newNum = self._origNum;

    var sigfig = new WILL_SignificantFigures(numStr); //WILL: TODO: possible optimisation - only generate these if needed
    self._origNumSigFigs = sigfig.sigFigs;
    self._origNumSigDecs = sigfig.sigDecs;
  }


  // helpers functions for dectectSeperators 
  function setThousandsStyleBlankUK() {
    self._thousandsSeperator = self.BLANK;
    self._decimalSeperator   = self.POINT;
  }
  function setThousandsStyleCommaUK() {
    self._thousandsSeperator = self.COMMA;
    self._decimalSeperator   = self.POINT;
  }
  function setThousandsStyleBlankEU() {
    self._thousandsSeperator = self.BLANK;
    self._decimalSeperator   = self.COMMA;
  }
  function setThousandsStylePointEU() {
    self._thousandsSeperator = self.POINT;
    self._decimalSeperator   = self.COMMA;
  }
  function setThousandsStyleQuote() {
    self._thousandsSeperator = self.QUOTE;
    self._decimalSeperator   = self.POINT;
  }
  function setThousandsStyleSpaceSwiss(decSep) {
    self._thousandsSeperator = self.SPACE;
    self._decimalSeperator   = decSep;
  }
  

  // sets self._thousandsSeperator, self._decimalSeperator
  function detectSeperators(number_str) {
    // Try and guess how this number is formatted
    var commas=0;
    var points=0;
    var spaces=0;
    var quotes=0;
    var last_symbol=null;

    for (var i=0;i<number_str.length;++i) {
      var c=number_str[i];
      if (c==self.COMMA ) { ++commas; }
      if (c==self.POINT ) { ++points; }
      if (c==self.SPACE ) { ++spaces; }
      if (c==self.QUOTE ) { ++quotes; }
    }
    
    if (spaces>0) {
      if         (commas==1 && !points) {
	return setThousandsStyleSpaceSwiss(self.COMMA); // spaces seperated thousands, decimal seperator=COMMA
      } else  if (points==1 && !commas) {
	return setThousandsStyleSpaceSwiss(self.POINT); // spaces seperated thousands, decimal seperator=POINT
      } else  if (!commas && !points) {
	return setThousandsStyleSpaceSwiss(self.POINT); // spaces seperated thousands, decimal seperator=POINT //WILL: POINT is an arbitary choice!
      }
      throw 'Found number formatted with spaces but cannot parse ['+number_str+']';
    }

    var m=number_str.match(/([\.\,])\d*$/);
    if (m) { last_symbol=m[1]; }

    if (quotes>=1) {
      if (spaces || commas || points >1) {
	throw 'Badly formatted number. If quote is used then  no SPACES, COMMAS  and maximum of one POINT. number='+number_str;
      }
      return setThousandsStyleQuote();
    }
  
    if (commas===0 && points===0) { //there is no formatting in this number. Let default to the point symbol.
      return setThousandsStyleBlankUK();
    }

    
    if (commas>1  && points>1 ) { throw 'Badly formatted number!  like 123.234.456,346,451 number='+number_str; }
    if (commas==1 && points==1) { //WILL: No error detection for group size
      if (last_symbol==self.POINT) {
	return setThousandsStyleCommaUK();
      } else {
	return setThousandsStylePointEU();
      }
    }
	
    if (commas==1 && points>1 ) {return setThousandsStylePointEU(); } //points must be the thousands seperator
    if (commas>1  && points==1) {return setThousandsStyleCommaUK(); } //commas must be the thousands seperator
    if ( (commas>0 && points===0) || (commas===0 && points>0) ){
      if (number_str[0]=='0') { //if leading digit is a 0: WILL: won't work for negative numbers!
	if (commas==1) {return setThousandsStyleBlankEU(); }
	if (points==1) {return setThousandsStyleBlankUK(); }
      }
      //we have 123.456.678 or 123,456,789
      //Does 123,456 mean hundreds or hundreds of thousands? --> We guess thousands due to grouping of 3
      var groups=number_str.split(/[\.,]/ , 10);
      var last_group=groups.pop(); //WILL: TODO: error check that all groups (apart from first) are of size 3?
      if (last_group.length==3) {
	//we found the thousands seperator - we return the decimal seperator
	if (last_symbol==self.POINT) {return setThousandsStylePointEU(); }
	if (last_symbol==self.COMMA) {return setThousandsStyleCommaUK(); }
      } else {
	//there was not exactly 3 digits in the last group, so it probably wasn't the thousands seperator
	if (last_symbol==self.POINT) {return setThousandsStyleBlankUK(); }
	if (last_symbol==self.COMMA) {return setThousandsStyleBlankEU(); }
      }
    }
    throw 'problem with '+number_str;
    //WILL: JSlint does not like: return undefined;
  }

  return this;  
}

WILL_Number.prototype = {
  
  // public funcs
  multiply : function(rate) {
    if (!rate) { throw "Invalid rate"; }
    this._newNum = this._origNum * rate; // Note that it's always the original number that is multiplied
  },

  // takes '1234567.89' --> '1,234,567.89' 
  // WILL: return string representation of number (currently only in UK/US format)
  // WILL: doesn't work with negative numbers!  
  toString: function() {
    var newNum=this._newNum;
    if (newNum===undefined || newNum===null) { throw "Invalid newNumber"; }
    if (newNum===NaN) { return "???"; }
    if (newNum===0) { return "0"; }
    if (newNum<0.000001) { return "???"; } // deals with too small and negative numbers
   
    var thousandsSeperator = this.COMMA; //WILL: hard coded
    var decimalSeperator   = this.POINT; //WILL: hard coded

    var numberStrJavascript = newNum.toString();

    var newNumberOfSigFigs  = this._origNumSigFigs;
    var newMaxDecimals      = -10; //this._origNumSigDecs;

    if (this._origNumSigDecs == -2) {
      // special case - keep format with 2 deciaml places $2.34 --> $5.67 - looks better
      newMaxDecimals=-2;
    }
      
    var dirty = (newNum != this._origNum);
    if (dirty) {
      ++newNumberOfSigFigs;
    }

    var numberStr;
    var r;
    var num;
    var fract;
     
    // format with thousands seperators. (WILL: TODO: merge with displaySigFigs)
   
    // first check the number of decimal places - if we are dirty (changed) and there is a fractional part and it's 1 digit long then increment its length
    do {
      numberStr=displaySigFigs(numberStrJavascript,newNumberOfSigFigs,newMaxDecimals,false); //scientific notation = false 
      
      r=numberStr.split(this.POINT,2);
      num=r[0];
      fract=r[1];
      ++newNumberOfSigFigs;
    } while ( dirty && fract && fract.length==1 );


    var c=0;
    var str='';
    for (var i=num.length-1;i>0;i--) {
      ++c;
      str=num[i]+str;
      if (c%3===0) {
	str=thousandsSeperator+str;
      }
    }
    str=num[0]+str; //add the first digit (it can't be a thousandsSeperator)
    if (fract) { str=str+decimalSeperator+fract; } //add the fractional part if it exists
    return str;
  }

};

 

 
var currencyLeftRegExp2; //WILL: TODO: this is currently a global var set in WILL_ConvertText
var currencyRightRegExp2;


// var matchThousandsSpaceRE    = /\d{1,2}(?:\s\d\d\d)*(?:[,.]\d+)?/; // 1 or 2 digits (space 3 digits)* (comma|dot digits+)?
// var matchThousandsCommaRE    = /\d{1,2}(?:\,\d\d\d)*(?:\.\d+)?/;   // 1 or 2 digits (comma 3 digits)* (dot       digits+)?
// var matchThousandsPointRE    = /\d{1,2}(?:\.\d\d\d)*(?:\,\d+)?/;   // 1 or 2 digits (point 3 digits)* (comma     digits+)?
// var matchThousandsBlankRE  = /\d+(?:[,.]\d+)?/g;   // 1 or more digits (comma|dot digits+)? //NB global

var TEXT='T'; //WILL: TODO: token types should be defined in OOP manner
var SPACE='W';
var SYMBOL='S';
var NUM='N';

var WINDOW_SIZE=20;
var MAX_NUMBER_LEN=20;
// NB: this only tokenises around numbers!! Not every whitespace and currency symbol is tokenised. Only if it's around a number.
function tokeniseNumbers(origText) {
  var tokenStream = {};
  tokenStream.tokens=[];
  tokenStream.types=[];
  var numberStr;
  var numberStartIndex;
  var numberEndIndex=0;
  var previousNumberEndIndex=0;
  var r; //for regexp results
  var toSearchStr; // window of text for looking for expanded number
  var leftStr; // text since previous number
  var firstIteration=true;
  // RegExp.input=origText; //WILL: TODO: can we optimise based on this?
  var matchDigit_RE = /\d/g;
  WILL_log('DBG3','looking for digits in ['+origText+']');
  while ( (r=matchDigit_RE.exec(origText)) ) { //find a digit 
    //WILL_log('DBG','Found a digit at '+r[0]);
    numberEndIndex=matchDigit_RE.lastIndex;
    numberStartIndex=numberEndIndex-1;

    // we limit the search area - WILL: TODO: is this necesary - maybe just search right context??
    toSearchStr=origText.substr(numberStartIndex,MAX_NUMBER_LEN);

    numberStr = WILL_findLongestNumber(toSearchStr);

    numberEndIndex=numberStartIndex+numberStr.length;
    matchDigit_RE.lastIndex=numberEndIndex; // manually move the next RegExp starting point since we may have found a longer number 
    leftStr = origText.slice(previousNumberEndIndex,numberStartIndex);
    //debug('left=['+leftStr+']');
    //debug('number=['+numberStr+']');

    tokeniseNonNumbers(tokenStream,leftStr,firstIteration);

    tokenStream.tokens.push(numberStr);
    tokenStream.types.push(NUM);

    previousNumberEndIndex=numberEndIndex;
    firstIteration=false;
  }
  if (numberEndIndex<origText.length) { // process any remaining text
    tokeniseNonNumbers(tokenStream,origText.slice(numberEndIndex));
  }


  // run through the tokens, if there is a number beside text (not a symbol, then mark it as text too)
  // exmaples: 3d or d3 - 3 is not a number here
  var tokens=tokenStream.tokens;
  var types=tokenStream.types;
  var i;
  //WILL: TODO: make tokenStream an object and this a method of it
  var findToken = function(type,offset) {
    var index=i+offset;
    if (index<0 || index>=types.length) { return false; }
    return (types[i+offset]==type);
  };
  
  var numJoinedToText;
  var matchBoundary = WILL_CurrencyList.matchBoundary;
  var matchStartBoundaryRegExp = new RegExp('^'+matchBoundary);
  var matchEndBoundaryRegExp   = new RegExp(matchBoundary+'$');
  for (i=0;i<tokens.length;i++) {
    if (findToken(NUM,0)) {
      numJoinedToText=false;
      if (findToken(TEXT,-1)) {
	if (!matchEndBoundaryRegExp.test(tokens[i-1])) { numJoinedToText=true; } // if the text token doesn't end with boundary char
      }
      //WILL: this breaks things like $55bn    - REMOVED
//       if (findToken(TEXT,+1)) {
// 	if (!matchStartBoundaryRegExp.test(tokens[i+1])) { numJoinedToText=true; } // if the text token doesn't start with boundary char
//       }
      if (numJoinedToText) {
	types[i]=TEXT; //change the number into text
      }
    }
  }

  var debugStr='';
  for (i=0;i<tokens.length;i++) {
    if (findToken(NUM,0)) { debugStr += tokens[i]+', '; }
  }
  if (debugStr) {
    WILL_log('DBG','Found numbers: '+debugStr);
  } else {
    WILL_log('DBG','Found no numbers');
  }
  
  return tokenStream;
}
  
function pushToken(stream,type,value) {
  stream.tokens.push(value);
  stream.types.push(type);
}

// search start and end of 'str' for whitespace and currency symbols //NB doesn't search middle if it's a big string
// - inserts TEXT, SYMBOL and SPACE tokens
function tokeniseNonNumbers(tokenStream,str,firstIteration) {
  if (!str) { return; }
  //debug('tokenising ['+str+']');
  var r;
  var spaceStr; //whitespace
  var matchedStr, boundaryStr, currencyStr;

  if (!firstIteration) {
    //    var leftWindow = str.substr(0,WINDOW_SIZE); //WILL: TODO: optimise by only searching a substring?
    r = currencyRightRegExp2.exec(str); // currency symbol to the right of digits e.g. 1 USD
    if (r) {
      spaceStr    =r[1]; //space between symbol and digit
      currencyStr =r[2]; //currency symbol
      boundaryStr =r[3]; //char before start of currency symbol
      
      if (spaceStr) { pushToken(tokenStream,SPACE,spaceStr); }
      pushToken(tokenStream,SYMBOL,currencyStr);     
      str = boundaryStr + RegExp.rightContext;
    }
    if (!str) { return; }
  }

  // IGNORE the middle part

  r = currencyLeftRegExp2.exec(str); // currency symbol is to the left of the digits e.g. $1
  if (r) {
    boundaryStr =r[1]; //char before start of currency symbol
    currencyStr =r[2]; //currency symbol
    spaceStr    =r[3]; //space between symbol and digit

    str = RegExp.leftContext + boundaryStr;

    if (str) { pushToken(tokenStream,TEXT,str); }
    pushToken(tokenStream,SYMBOL,currencyStr);
    if (spaceStr) { pushToken(tokenStream,SPACE,spaceStr); }

      
  } else {
    // there was no currency in the right hand side - so push the remaining str as TEXT or WHITESPACE
    if ( /\S/.test(str)) { pushToken(tokenStream,TEXT,str); }
    else {                 pushToken(tokenStream,SPACE,str); }
  }

}

    
function associateNumbers(tokenStream) {

  var tokens=tokenStream.tokens;
  var types =tokenStream.types;
  var symbol2number=[];
  var number2symbol=[]; // [numberPos,symbolPos,spaceChars]
  var space;
  var i;

  var findToken = function(type,offset) {
    var index=i+offset;
    if (index<0 || index>=types.length) { return false; }
    return (types[i+offset]==type);
  };
  var findTokenSpaceLen = function(offset) {
    if (findToken(SPACE,offset)) { return tokens[i+offset].length; }
    else { return null; }
  };

  //for (i=0;i<tokens.length;i++) { number2symbol[i]=[]; }

  for (i=0;i<tokens.length;i++) {
    //debug('token:'+tokens[i]+' '+types[i]);
    if (!findToken(NUM,0)) { continue; } //skip non-numbers

    for (var dir=-1; dir<=1; dir+=2) { //look before and after current number
      if (findToken(SYMBOL,dir)) { number2symbol.push([i, (i+dir) , 0] ); }
      else {
	space = findTokenSpaceLen(dir);
	if (space && findToken(SYMBOL,dir*2)) { number2symbol.push([ i, (i+dir*2) , space]); }
      }
    }
  }

  for (var sortByIndex=0; sortByIndex<=1; sortByIndex++) {
    // find the shortest link for each num
    var minDist=[];
    var num;
    var dist;
    for (i=0;i<number2symbol.length;i++) {
      num=number2symbol[i][sortByIndex];      dist=number2symbol[i][2];
      if (minDist[num]===undefined || minDist[num]>dist) { minDist[num]=dist; }
    }
      
    var dups=[];
    // mark all but the shortest as unused, check for duplicates
    for (i=0;i<number2symbol.length;i++) {
      num=number2symbol[i][sortByIndex];   dist=number2symbol[i][2];
      if (!dups[num]) {dups[num]=0;} //initialise hashtable
	
      if (dist != minDist[num]) { number2symbol[i][3]=true; }
      else                      { dups[num]++; }
    }
      
    for (i=0;i<number2symbol.length;i++) {
      num=number2symbol[i][sortByIndex];
      if (dups[num]>1) { // we found 2 links at minimum distance. This number is ambigious. Mark as unused
	number2symbol[i][3]=true;
      }
    }
  }

  var number2symbolOK=[];
  for (i=0;i<number2symbol.length;i++) {
    if (!number2symbol[i][3]) {
      number2symbolOK.push( number2symbol[i][0]);
      number2symbolOK.push( number2symbol[i][1]);
    }
  }

  //     for (i=0;i<number2symbolOK.length;i++) {
  //       info('number2symbolOK['+i+']='+number2symbolOK[i]);
  //     }

  return number2symbolOK;

}
//WILL: TODO: DOCS: symbol means currencySymbol
function WILL_CurrencyConvertStr(origText, targetCurrency, sourceCurrencyList, getRateFunc, hideOriginal, defaultSourceCurr) { //the new version
  //info('NEW_CurrencyConvertStr: ['+origText+']');
  currencyLeftRegExp2 = sourceCurrencyList.getCurrencyLeftRegExp2(); //setup the global vars used by tokenise
  currencyRightRegExp2 = sourceCurrencyList.getCurrencyRightRegExp2();
  var tokenStream=tokeniseNumbers(origText);
  var tokens=tokenStream.tokens;
  var types=tokenStream.types;

  //hashtable of numberPositions that have already been converted. Used to convert any numbers without symbols to default curr
  var numberWasConverted=[]; 

  if (!targetCurrency) { throw('missing target currency'); }

  var numberAssociation = associateNumbers(tokenStream);    
  //info('numberAssociation='+numberAssociation);
  var numberPos;
  var symbolPos;
  var numberStr;
  var symbolStr;
  var firstPos;
  var secondPos;
  var firstStr;
  var secondStr;
  var originalStr;
  var spaceBetween;
  var newNumberStr;
  var numberBeforeSymbol;
  var rate;
  var sourceCurrency;
  var spaceStr;
  
  for (var i=0;i<numberAssociation.length;i+=2) {
    numberPos=numberAssociation[i+0];
    symbolPos=numberAssociation[i+1];
    numberStr=tokens[numberPos];
    symbolStr=tokens[symbolPos];
    //info('converting '+symbolStr+numberStr);

    sourceCurrency=sourceCurrencyList.findFromStr(symbolStr,true); //fail silently
    if (!sourceCurrency) { 
      WILL_log('WARN','Currency is recongised in the RegExp list, but it is not in the sourceCurrencyList');
      continue; //skip this one
    }

    numberWasConverted[numberPos]=true; // if we recognised this as a money then mark it as converted. It should not be considered for default conversion

    if (sourceCurrency.equals(targetCurrency)) { continue; } // skip this money, it's already in the target currency

    firstPos  = Math.min(symbolPos,numberPos);
    secondPos = Math.max(symbolPos,numberPos);
    if      ((secondPos-firstPos)==1) { spaceBetween=false; }
    else if ((secondPos-firstPos)==2) { spaceBetween=true;  }
    else { throw 'WILL: number and symbol are not 1 or 2 tokens apart. numberPos='+numberPos+' symbolPos='+symbolPos; }

    if (spaceBetween) { spaceStr=' '; } //this only used for (original_value)
    else              { spaceStr='';  }

    numberBeforeSymbol = (numberPos==firstPos); // is number first or second?

    rate=getRateFunc(sourceCurrency,targetCurrency);
    if (!rate) { return origText; } //WILL: TODO: throw('no rate found: WILL: TODO: handle this better!'); 
    var number = new WILL_Number(numberStr);
    number.multiply(rate);
    newNumberStr=number.toString();
      
    if (numberBeforeSymbol) {
      firstStr = newNumberStr;
      secondStr = targetCurrency.getHtml();
      originalStr = ' ('+numberStr+spaceStr+symbolStr+')';
    } else {
      firstStr  = targetCurrency.getHtml();
      secondStr = newNumberStr;
      originalStr = ' ('+symbolStr+spaceStr+numberStr+')';
    }
    if (hideOriginal) { originalStr=''; }
    tokens[firstPos]  = firstStr;
    tokens[secondPos] = secondStr + originalStr;

  }

  if (defaultSourceCurr) {
    // go through and find unconverted numbers and use default source currency to convert them
    for (i=0;i<tokens.length;i++) {
      if (types[i]==NUM && !numberWasConverted[i]) {
	numberStr = tokens[i];
	rate=getRateFunc(defaultSourceCurr,targetCurrency);
	if (!rate) { throw('no rate found: WILL: TODO: handle this better!'); }
	number = new WILL_Number(numberStr);
	number.multiply(rate);
	newNumberStr=number.toString();
	if (hideOriginal) { originalStr=''; }
	else              { originalStr = ' ('+defaultSourceCurr.getHtml()+numberStr+')'; }
	tokens[i]=targetCurrency.getHtml()+newNumberStr+originalStr;
	// (we do not add spaces between currency and number to minimise the extra chars we are adding)

	// don't need to mark numberWasConverted
      }
    }
  }

  //info('NEW_CurrencyConvertStr --> ['+tokens.join('')+']');
  return tokens.join('');

}
