// $Id: WILL_PrefManager.js,v 1.15 2006/01/05 10:19:49 Will Exp $

// Wrapper around the preferences-service. Inspired by GreaseMonkey
function WILL_PrefManager(startpoint) {
  var self=this;

  var STARTPOINT = startpoint || 'extensions.viewmycurrency.';  // default normally used
  
  // var KEY_LAST_UPDATE             = 'fx.last_update'; //fx = Foreign eXchange
  var KEY_RATE                    = 'fx.rate.';
  var KEY_CURRENT_VERSION         = 'currentversion';

  var KEY_SOURCE_CURRENCY_LIST    = 'list.sourceCurrencyList';
  var KEY_COMPLETE_CURRENCY_LIST  = 'list.completeCurrencyList';
  var KEY_TARGET_CURRENCY_LIST    = 'list.targetCurrencyList';

  var KEY_TARGET_CURRENCY         = 'target_currency';
  var KEY_HIDE_ORIGINAL_CURRECY   = 'hide_original_currency';
  var KEY_DEBUG = 'debug';
  var KEY_WEBSERVICE_URL = 'webserviceUrl';

  this.requiredPrefs=[KEY_TARGET_CURRENCY,KEY_SOURCE_CURRENCY_LIST,KEY_COMPLETE_CURRENCY_LIST,KEY_TARGET_CURRENCY_LIST,KEY_WEBSERVICE_URL+'.1'];
  this.optionalPrefs=[KEY_HIDE_ORIGINAL_CURRECY,KEY_DEBUG];

  var pref = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch(STARTPOINT);

  var SOURCE_CURRENCY_LIST_CODES=['USD','EUR','GBP','JPY','CAD','AUD','CHF','CNY'];
  var TARGET_CURRENCY_LIST_CODES=['USD','EUR'];
  var COMPLETE_CURRENCY_LIST = WILL_createCompleteCurrencyList();
  
  this.observers = {};
  this.func2pref = {};

  var EARLIEST_COMPATIBLE_VERSION = 0.25;

  //////////////////////
  // public functions //
  //////////////////////
  
  // on application startup we do these checks but we don't change anything (autoCancel)
  // once we know the status, we do the checks again when the user clicks on me.
  this.doStartupChecks = function(autoCancel) {
    var ALREADY_INSTALLED = "ALREADY_INSTALLED"; // compatible version is already installed (normal case on startup!)
    var DONT_INSTALL="DONT_INSTALL"; // invalid prefs --> first time (missing or old version num) or (user didn't want to delete old prefs and start again)
    var TO_INSTALL = "TO_INSTALL";  // first time (everything empty) or either case of DONT_INSTALL but user anwswers OK.

    // Do the version check
    if (!this.exists(KEY_CURRENT_VERSION)) {   // we are missing the version number
      // but maybe there is some old junk lying  around?
      if (!this._checkPromptUserAndDeleteBranch(autoCancel,STARTPOINT)) { return DONT_INSTALL; }
      else { return TO_INSTALL; } // to install (no version num)
    }

    //version is key present

    var storedVersion = Number(this.getString(KEY_CURRENT_VERSION)).valueOf();
    if (storedVersion < EARLIEST_COMPATIBLE_VERSION) {
      if (!this._checkPromptUserAndDeleteBranch(autoCancel,STARTPOINT)) { return DONT_INSTALL; }
      else { return TO_INSTALL; } // to re-install (out of date)
    }

    // version was up to date
    
    var missingReqPrefs=this.missingRequiredPrefs();
    if (missingReqPrefs.length) {
      if (!this._checkPromptUserAndDeleteBranch(autoCancel,STARTPOINT,"Required ViewMyCurrency preferences have been deleted.")) { return DONT_INSTALL; }
      else { return TO_INSTALL; } // to re-install (missing prefs)
    }

    // check if any of the optional prefs are missing, and if so, restore them
    var missingOptPrefs=this.missingOptionalPrefs();
    if (missingOptPrefs) { this.restoreOptionalPrefs(); }

    // version was up to date. No further setup action requred
    return ALREADY_INSTALLED;
  };
   
  this.doFirstTimeSetup = function(defaultTargetCurrencyList) {

    var newTargetCurrencyList; //set either from arguemnt or from dialog box
    if (defaultTargetCurrencyList) { // currently only set during testing
      newTargetCurrencyList=defaultTargetCurrencyList;
    } else {
      var retVal       = {Message:'Return values go here'};
      window.openDialog( 'chrome://viewmycurrency/content/WILL_StartUpDialog.xul','WILL_StartUpDialog','chrome,resizable,modal',
			 COMPLETE_CURRENCY_LIST,this.getDefaultSourceListPref(),retVal );
      //WILL_dbg(retVal);
      if (retVal['newTargetCurrency']) {
	newTargetCurrencyList=new WILL_CurrencyList(retVal['newTargetCurrency']);
      } else {
	return false; // the user clicked cancel
      }
    }
    
    // The click finished the setup wizard. Store all the prefs
    this.setString(KEY_CURRENT_VERSION,VIEWMYCURRENCY_VERSION);
    this.restoreDefaultPrefs();
    this.restoreOptionalPrefs();

    if (newTargetCurrencyList) {
      this.setTargetCurrencyList( newTargetCurrencyList );
      this.setTargetCurrency( newTargetCurrencyList.items[0] ); //start converting
    }
    // else we just use the default settings

    return true; // we are now setup
  };

  // we seperate required and optional prefs, missing optional prefs can be restored without asking the user
  this.restoreOptionalPrefs = function() {
    this.setDebugLevel(0);               // no debug
    this.setHideOriginalCurrency(false); // no hide
  };
  
  this.getDefaultSourceListPref   = function() { return new WILL_CurrencyList(COMPLETE_CURRENCY_LIST.findFromCodes(SOURCE_CURRENCY_LIST_CODES)); };
  this.getDefaultTargetListPref   = function() { return new WILL_CurrencyList(COMPLETE_CURRENCY_LIST.findFromCodes(TARGET_CURRENCY_LIST_CODES)); };
  this.getDefaultCompleteListPref = function() { return WILL_createCompleteCurrencyList(); };

  this.restoreDefaultPrefs = function() {
    this.setSourceCurrencyList(   this.getDefaultSourceListPref()   );
    this.setTargetCurrencyList(   this.getDefaultTargetListPref()   );
    this.setCompleteCurrencyList( this.getDefaultCompleteListPref() ); 
    // this.setWebserviceUrl(1,'chrome://viewmycurrency/content/tests/fixed_rates/FROM_CODE2TO_CODE');   
    this.setWebserviceUrl(1,'http://www.webservicex.net/CurrencyConvertor.asmx/ConversionRate?FromCurrency=FROM_CODE&ToCurrency=TO_CODE');
    this.setTargetCurrency(null); //disabled by default
    //    this.resetLastUpdateAttemptTime();
  };

  this.watchDebugLevel = function(context,func) { this.addWatcher(KEY_DEBUG,                context,func); };
  this.watchTargetList = function(context,func) { this.addWatcher(KEY_TARGET_CURRENCY_LIST, context,func); };
  this.watchSourceList = function(context,func) { this.addWatcher(KEY_SOURCE_CURRENCY_LIST, context,func); };
  this.watchHideOriginal=function(context,func) { this.addWatcher(KEY_HIDE_ORIGINAL_CURRECY,context,func); };
  this.getDebugLevel = function() {

    //if key exists and contains anything
    if (this.exists(KEY_DEBUG)) {
      var level = this.getString(KEY_DEBUG);
      if (level=='0') { return 0; }
      if (level=='1') { return 1; }
      if (level=='2') { return 2; }
      if (level=='3') { return 3; }
      WILL_log('ERR','Unknown level for debug '+level);
      return 0;
    }
    return 0;
  };

  this.setDebugLevel = function(level) {
    this.setString(KEY_DEBUG,level);
  };

  this.getHideOriginalCurrency = function() {
    //if key exists and contains anything
    if (this.exists(KEY_HIDE_ORIGINAL_CURRECY)) {
      if (this.getString(KEY_HIDE_ORIGINAL_CURRECY)) { return true; }
    }

    return false;
  };

  this.setHideOriginalCurrency = function(enabled) {
    if (enabled) {
      this.setString(KEY_HIDE_ORIGINAL_CURRECY,"enabled");
    } else {
      this.setString(KEY_HIDE_ORIGINAL_CURRECY,'');
    }
  };

  this.setSourceCurrencyList   = function(list) { this.setCurrencyList(KEY_SOURCE_CURRENCY_LIST,list);    };
  this.setTargetCurrencyList   = function(list) { this.setCurrencyList(KEY_TARGET_CURRENCY_LIST,list);    };
  this.setCompleteCurrencyList = function(list) { this.setCurrencyList(KEY_COMPLETE_CURRENCY_LIST,list);  };

  this.getSourceCurrencyList   = function()     { return this.getCurrencyList(KEY_SOURCE_CURRENCY_LIST);   };
  this.getTargetCurrencyList   = function()     { return this.getCurrencyList(KEY_TARGET_CURRENCY_LIST);   };
  this.getCompleteCurrencyList = function()     { return this.getCurrencyList(KEY_COMPLETE_CURRENCY_LIST); };

  this.setCurrencyList = function(key,currencyList) { this.setString(key,currencyList.toString(true));   };
  this.getCurrencyList = function(key)              { return new WILL_CurrencyList(this.getString(key)); };

  this.missingRequiredPrefs = function () {
    var missing='';
    var key;
    for (var i=0; i<this.requiredPrefs.length; i++) {
      key=this.requiredPrefs[i];
      if (!this.exists(key)) { missing=missing+key+','; }
    }
    return missing;
  };

  this.missingOptionalPrefs = function () {
    var missing='';
    var key;
    for (var i=0; i<this.optionalPrefs.length; i++) {
      key=this.optionalPrefs[i];
      if (!this.exists(key)) { missing=missing+key+','; }
    }
    return missing;
  };

  this.getTargetCurrency = function () {
    var code=this.getString(KEY_TARGET_CURRENCY);
    if (!code) { return null; } //null is possible (means disabled)
    var curr=COMPLETE_CURRENCY_LIST.findFromCode(code); 
    return curr; 
  };

  this.setTargetCurrency = function (newTarget) {
    var code=''; //disabled code
    if (newTarget) { code=newTarget.code; }
    this.setString(KEY_TARGET_CURRENCY,code);
  };

//   this.resetLastUpdateAttemptTime = function() { this.setString(KEY_LAST_UPDATE,"");  };
//   this.setLastUpdateAttemptTime   = function() { this.setString(KEY_LAST_UPDATE,new Date().toString());  }; //Time-stamp the update
//   this.getLastUpdateAttemptTime   = function() { return this.getString(KEY_LAST_UPDATE);  };

  //WILL: must use 'self' because this function is called by ExchangeRateCache ... WILL: TODO: is this a good idea?
  this.setWebserviceUrl = function(num,url) { this.setString(KEY_WEBSERVICE_URL+'.'+num, url);   };
  this.getWebserviceUrl = function(num)     { return self.getString(KEY_WEBSERVICE_URL+'.'+num); };

  this.exists = function(key) {
    var prefType = pref.getPrefType(key);
    if (prefType == pref.PREF_INVALID) { return false; }
    else                               { return true;  }
  };
  
  this.readRate = function(rateObj,failSilently) {
    var key=KEY_RATE+rateObj.toCurrency.code+'/'+rateObj.fromCurrency.code;
    if (!self.exists(key)) {
      if (!failSilently) { WILL_log('WARN','PrefManager: missing '+key); }
      return false;
    }
    rateObj.updateFromStr(self.getString(key));
    WILL_log('DBG','PrefManager: retrieved '+rateObj);
    return true;
  };

  this.writeRate = function(rateObj) {
    var key=KEY_RATE+rateObj.toCurrency.code+'/'+rateObj.fromCurrency.code;
    self.setString(key,rateObj.toString(true)); //store encoded rate object
  };


  // The setter
  this.setString = function(key,str) {
    WILL_log('DBG','PrefManager: Storing '+key+'='+str);
    pref.setCharPref(key, str);
  };
   
  // The getter
  this.getString = function(key) {
    if (!this.exists(key)) {
      WILL_log('ERR','PrefManager: pref does not exist ['+key+']');
      return null;
    }
    return pref.getCharPref(key);
  };



    
  //WILL: TODO: Credit Greasemonkey
  /**
   * call a function whenever the named preference subtree changes
   */
  this.addWatcher = function(prefName, contextObj,watcherFuncStr) {
    // construct an observer
    var observer = {
      observe:function(subject, topic, prefName) {
	var callwrapper = new CCallWrapper(contextObj, 10, watcherFuncStr, prefName); //after 10 ms delay call contextObj.'watcherFuncStr'(prefName)
	CCallWrapper.asyncExecute(callwrapper);
	//watcherFuncStr.call(contextObj,prefName);
      }
    };
    
    // store the observer in case we need to remove it later
    this.observers[watcherFuncStr] = observer;
    this.func2pref[watcherFuncStr] = prefName;
    WILL_log('DBG','Watching '+prefName);
    pref.QueryInterface(Components.interfaces.nsIPrefBranchInternal).addObserver(prefName, observer, false);
  };
  
  /**
   * stop watching
   */
  this.removeWatcher = function(watcherFuncStr) {
    var prefName=this.func2pref[watcherFuncStr];
    if (this.observers[watcherFuncStr]) {
      pref.QueryInterface(Components.interfaces.nsIPrefBranchInternal).removeObserver(prefName, this.observers[watcherFuncStr]);
    } else {
      WILL_alert('tried to remove unknown watcherFuncStr for pref '+prefName);
    }
  };



  this._checkPromptUserAndDeleteBranch = function(autoCancel,startpoint,messageStart,autoOK) {
    var root = Components.classes["@mozilla.org/preferences-service;1"]. getService(Components.interfaces.nsIPrefBranch);
    var children = root.getChildList(startpoint, {});
    if (!children.length) { return true;  } //branch doesn't exist
    if (autoCancel)       { return false; } //branch exists and has not been deleted

    var message= messageStart || 'Incompatible preferences from an older version of ViewMyCurrency have been detected.\n';
    message=message+ '\nDelete the '+children.length+' items stored in "'+startpoint+'" ?';

    var result;
    if (!autoOK) {
      var prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
      result = prompts.confirm(window, "Delete old preferences", message);
    }

    if (result || autoOK) {
      //alert('deleting '+children.join(','));
      root.deleteBranch(startpoint);
      return true; //they existed but now they are gone
    } else {
      return false; //user clicked cancel - prefs exist but not deleted
    }

  };

  this.deleteViewMyCurrencyPrefs = function() {
    var root = Components.classes["@mozilla.org/preferences-service;1"]. getService(Components.interfaces.nsIPrefBranch);    
    var prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
    var result = prompts.confirm(window, "Delete all ViewMyCurrency prefs?","Do you want to delete all preferences associated with ViewMyCurrency");
    if (result) {
      root.deleteBranch(STARTPOINT);
    }
  };

}

